# SPDX-License-Identifier: MPL-2.0
"""IKEv2 message parsers."""
# pylint: disable=too-many-lines

import abc
import collections
import enum
import typing

import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.ike.algorithm import (
    Ikev2TransformAttributeType,
    Ikev2PayloadType,
    Ikev2DiffieHellmanGroup,
    Ikev2EncryptionAlgorithm,
    Ikev2IntegrityAlgorithm,
    Ikev2NotifyType,
    Ikev2ProtocolId,
    Ikev2PseudorandomFunction,
    Ikev2TransformType,
)
from cryptoparser.common.base import TwoByteEnumParsable, VariantParsable
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData, TooMuchData, InvalidType
from cryptoparser.ike.common import DataAttributeLength, DataAttributeTypeValue, DataAttributeFormat


class Ikev2ProposalFlags(enum.IntFlag):
    """Proposal flags."""
    LAST_SUBSTRUCT = 0x80


class Ikev2PayloadFlags(enum.IntFlag):
    """Payload flags."""
    CRITICAL = 0x80  # Critical bit flag


@attr.s
class Ikev2PayloadBase(ParsableBase):
    """Payload header parser, according to RFC7296.

    The generic payload header has the following structure:

    .. code-block:: text

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next Payload  |C|  RESERVED   |         Payload Length        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :cvar HEADER_SIZE: Size of the header in bytes
    :ivar next_payload: Type of the next payload (1 byte)
    :ivar critical: Critical bit flag (1 bit)
    """
    HEADER_SIZE = 4

    flags: typing.Set[Ikev2PayloadFlags] = attr.ib(
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(Ikev2PayloadFlags),
        )
    )
    next_payload: typing.Optional[Ikev2PayloadType] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(Ikev2PayloadType))
    )

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_payload_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        """Parse payload header from bytes.

        :param parsable: Bytes to parse
        :type parsable: bytes
        :return: Tuple of (parsed header, number of bytes parsed)
        :rtype: tuple(PayloadBase, int)
        :raises NotEnoughData: If there are not enough bytes to parse
        """
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_numeric_enum_coded('next_payload', Ikev2PayloadType)
        parser.parse_numeric_flags('flags', 1, Ikev2PayloadFlags)
        parser.parse_numeric('payload_length', 2)

        if parser.unparsed_length < parser['payload_length'] - cls.HEADER_SIZE:
            raise NotEnoughData(parser['payload_length'] - cls.HEADER_SIZE - parser.unparsed_length)

        return parser

    def compose_header(self, payload_length):
        """Compose payload header to bytes.

        :return: Composed header bytes
        :rtype: bytes
        """
        assert self.next_payload is not None

        composer = ComposerBinary()
        composer.compose_numeric(self.next_payload.value.code, 1)
        composer.compose_numeric_flags(self.flags, 1)
        composer.compose_numeric(self.HEADER_SIZE + payload_length, 2)

        return composer


class TransformNextPayload(enum.IntEnum):
    """Transform next payload."""
    LAST = 0x00
    MORE = 0x03


@attr.s
class Transform(ParsableBase):
    """Transform payload parser.

    The transform payload has the following format:

    .. code-block:: text

                             1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next payload  |   RESERVED    |        Transform Length       |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |Transform Type |   RESERVED    |          Transform ID         |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        ~                      Transform Attributes                     ~
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar transform_id: Transform ID
    :ivar next_payload: Next payload
    """
    HEADER_SIZE = 8

    transform_id: typing.Any = attr.ib()
    next_payload: typing.Optional[TransformNextPayload] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(TransformNextPayload))
    )

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_transform_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_transform_id_class(cls):
        raise NotImplementedError()

    @transform_id.validator
    def _validate_transform_id(self, _, value):
        transform_id_class = self._get_transform_id_class()
        if not isinstance(value, transform_id_class):
            raise InvalidValue(value, type(self), 'value')

    @classmethod
    def _parse_header(cls, parsable):
        """Parse transform from bytes.

        :param parsable: Bytes to parse
        :type parsable: bytes
        :return: Tuple of (parsed transform, number of bytes parsed)
        :rtype: tuple(Transform, int)
        :raises NotEnoughData: If there are not enough bytes to parse
        """
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_numeric('next_payload', 1, TransformNextPayload)
        parser.parse_numeric('reserved', 1)
        parser.parse_numeric('transform_length', 2)
        parser.parse_numeric_enum_coded('transform_type', Ikev2TransformType)

        if parser['transform_type'] != cls.get_transform_type():
            raise InvalidType()

        parser.parse_numeric('reserved2', 1)
        parser.parse_numeric_enum_coded('transform_id', cls._get_transform_id_class())

        return parser

    def compose_header(self, transform_length):
        """Compose transform to bytes.

        :param last_substruc: Whether this is the last substructure
        :type last_substruc: bool
        :return: Composed transform bytes
        :rtype: bytes
        """
        assert self.next_payload is not None

        composer = ComposerBinary()
        composer.compose_numeric(self.next_payload.value, 1)
        composer.compose_numeric(0, 1)  # reserved
        composer.compose_numeric(transform_length + self.HEADER_SIZE, 2)
        composer.compose_numeric(self.get_transform_type().value.code, 1)
        composer.compose_numeric(0, 1)  # reserved2
        composer.compose_numeric(self.transform_id.value.code, 2)

        return composer


class TransformAttributeKeyLength(DataAttributeLength):
    @classmethod
    def get_type(cls):
        return Ikev2TransformAttributeType.KEY_LENGTH

    @classmethod
    def _get_size(cls):
        return 2


@attr.s
class TransformAttributeSignatureAlgorithm(DataAttributeTypeValue):
    """Signature Algorithm transform attribute (TLV format).

    The Signature Algorithm attribute has the following format:

    0                   1                   2                   3
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |A|       Attribute Type        |    Attribute Length            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                                                               |
    ~                   Signature Algorithm Value                   ~
    |                                                               |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar signature_algorithm: Signature algorithm value
    """
    signature_algorithm: typing.Union[bytearray, bytes] = attr.ib(
        validator=attr.validators.instance_of((bytearray, bytes))
    )

    @classmethod
    def get_type(cls):
        return Ikev2TransformAttributeType.SIGNATURE_ALGORITHM

    @classmethod
    def _get_format(cls):
        return DataAttributeFormat.TYPE_LENGTH_VALUE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        if parser.unparsed_length < 2:
            raise NotEnoughData(2 - parser.unparsed_length)

        parser.parse_numeric('length', 2)
        parser.parse_raw('signature_algorithm', parser['length'])

        return cls(signature_algorithm=bytes(parser['signature_algorithm'])), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_numeric(len(self.signature_algorithm), 2)
        composer.compose_raw(self.signature_algorithm)

        return composer.composed_bytes


@attr.s
class TransformNoAttributes(Transform):
    """Transform payload parser for transforms with no attributes."""
    @classmethod
    @abc.abstractmethod
    def get_transform_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_transform_id_class(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        transform = cls(
            transform_id=parser['transform_id'],
        )
        transform.next_payload = parser['next_payload']

        return transform, parser.parsed_length

    def compose(self):
        return self.compose_header(transform_length=0).composed_bytes


@attr.s
class TransformAttributes(Transform):
    """Transform payload parser for transforms with some attributes."""

    @classmethod
    @abc.abstractmethod
    def get_transform_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_transform_id_class(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_attributes(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def _get_attributes(self):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)
        attributes, attributes_length = cls._parse_attributes(header_parser.unparsed)

        transform = cls(
            transform_id=header_parser['transform_id'],
            **attributes
        )

        transform.next_payload = header_parser['next_payload']

        return transform, header_parser.parsed_length + attributes_length

    def compose(self):
        payload_composer = ComposerBinary()
        for attribute in self._get_attributes():
            payload_composer.compose_parsable(attribute)

        header_composer = self.compose_header(transform_length=payload_composer.composed_length)

        return header_composer.composed_bytes + payload_composer.composed_bytes


class Ikev2TransformIntegrity(TransformNoAttributes):
    """Transform payload parser for integrity algorithm."""

    @classmethod
    def get_transform_type(cls):
        return Ikev2TransformType.INTEG

    @classmethod
    def _get_transform_id_class(cls):
        return Ikev2IntegrityAlgorithm


class Ikev2TransformPrf(TransformNoAttributes):
    """Transform payload parser for pseudorandom function."""

    @classmethod
    def get_transform_type(cls):
        return Ikev2TransformType.PRF

    @classmethod
    def _get_transform_id_class(cls):
        return Ikev2PseudorandomFunction


class Ikev2TransformDhGroup(TransformNoAttributes):
    """Transform payload parser for Diffie-Hellman group."""

    @classmethod
    def get_transform_type(cls):
        return Ikev2TransformType.DH

    @classmethod
    def _get_transform_id_class(cls):
        return Ikev2DiffieHellmanGroup


@attr.s
class Ikev2TransformEncryptionAlgorithm(TransformAttributes):
    """Transform payload parser for encryption algorithm."""

    key_length: int = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_transform_type(cls):
        return Ikev2TransformType.ENCR

    @classmethod
    def _get_transform_id_class(cls):
        return Ikev2EncryptionAlgorithm

    @classmethod
    def _parse_attributes(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('key_length', TransformAttributeKeyLength)

        return {'key_length': parser['key_length'].value}, parser.parsed_length

    def _get_attributes(self):
        return [TransformAttributeKeyLength(value=self.key_length)]


class Ikev2ProposalNextPayload(enum.IntEnum):
    """Proposal next payload."""
    LAST = 0x00
    MORE = 0x02


@attr.s
class Ikev2Proposal(ParsableBase):
    """Proposal payload parser.

    The proposal payload has the following format:

    .. code-block:: text

                            1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Last Substruc |   RESERVED    |         Proposal Length       |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Proposal Num  |  Protocol ID  |    SPI Size   |Num  Transforms|
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       ~                        SPI (variable)                         ~
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                        <Transforms>                           ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar flags: Proposal flags (Ikev2ProposalFlags)
    :ivar protocol_id: Protocol ID
    :ivar spi: Security Parameter Index
    :ivar transforms: List of transforms
    """
    HEADER_SIZE = 8

    protocol_id: Ikev2ProtocolId = attr.ib(validator=attr.validators.instance_of(Ikev2ProtocolId))
    transforms: typing.List[Transform] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(Transform)
    ))
    spi: bytes = attr.ib(default=bytes(), converter=bytes, validator=attr.validators.instance_of(bytes))
    last: typing.Optional[Ikev2ProposalNextPayload] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(Ikev2ProposalNextPayload))
    )
    proposal_number: typing.Optional[int] = attr.ib(
        init=False,
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int))
    )

    @classmethod
    def _parse(cls, parsable):
        """Parse proposal from bytes.

        :param parsable: Bytes to parse
        :type parsable: bytes
        :return: Tuple of (parsed proposal, number of bytes parsed)
        :rtype: tuple(Ikev2ProposalPayload, int)
        :raises NotEnoughData: If there are not enough bytes to parse
        """
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_numeric_flags('last', 1, Ikev2ProposalNextPayload)
        parser.parse_numeric('reserved', 1)
        parser.parse_numeric('proposal_length', 2)
        parser.parse_numeric('proposal_number', 1)
        parser.parse_numeric_enum_coded('protocol_id', Ikev2ProtocolId)
        parser.parse_numeric('spi_size', 1)
        parser.parse_numeric('transform_count', 1)
        parser.parse_raw('spi', parser['spi_size'])

        transforms = []
        for _ in range(parser['transform_count']):
            parser.parse_parsable('transform', Ikev2TransformVariantInitiator)
            transforms.append(parser['transform'])

        proposal = cls(
            protocol_id=parser['protocol_id'],
            spi=parser['spi'],
            transforms=transforms
        )

        proposal.last = parser['last']
        proposal.proposal_number = parser['proposal_number']

        return proposal, parser.parsed_length

    def compose(self):
        """Compose proposal to bytes.

        :param last_substruc: Whether this is the last substructure
        :type last_substruc: bool
        :return: Composed proposal bytes
        :rtype: bytes
        """
        assert self.last is not None

        header_composer = ComposerBinary()
        header_composer.compose_numeric(self.last.value, 1)
        header_composer.compose_numeric(0, 1)  # reserved

        payload_composer = ComposerBinary()
        for i, transform in enumerate(self.transforms):
            transform.next_payload = (
                TransformNextPayload.MORE
                if i < len(self.transforms) - 1
                else TransformNextPayload.LAST
            )
            payload_composer.compose_parsable(transform)

        header_composer.compose_numeric(self.HEADER_SIZE + len(payload_composer.composed_bytes), 2)
        header_composer.compose_numeric(self.proposal_number, 1)
        header_composer.compose_numeric(self.protocol_id.value.code, 1)
        header_composer.compose_numeric(len(self.spi), 1)
        header_composer.compose_numeric(len(self.transforms), 1)
        if self.spi:
            header_composer.compose_raw(self.spi)

        return header_composer.composed_bytes + payload_composer.composed_bytes


@attr.s
class Ikev2PayloadSecurityAssociation(Ikev2PayloadBase):
    """Security Association payload parser.

    The Security Association payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                          <Proposals>                          ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """
    proposals: typing.List[Ikev2Proposal] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(Ikev2Proposal)
    ))

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.SA

    def get_transform_by_type(self, transform_type: Ikev2TransformType) -> Transform:
        for proposal in self.proposals:
            for transform in proposal.transforms:
                if transform.get_transform_type() == transform_type:
                    return transform

        raise KeyError(transform_type)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        proposals = []
        parser_proposal = ParserBinary(parser.unparsed[:parser['payload_length'] - cls.HEADER_SIZE:])
        while parser_proposal.unparsed_length > 0:
            parser_proposal.parse_parsable('proposal', Ikev2Proposal)
            proposal = parser_proposal['proposal']
            proposals.append(proposal)

        payload = cls(
            flags=parser['flags'],
            proposals=proposals
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length + parser_proposal.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()

        for proposal_number, proposal in enumerate(self.proposals):
            proposal.last = (
                Ikev2ProposalNextPayload.MORE
                if proposal_number < len(self.proposals) - 1
                else Ikev2ProposalNextPayload.LAST
            )
            proposal.proposal_number = proposal_number + 1
            composer_payload.compose_parsable(proposal)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev2PayloadNonce(Ikev2PayloadBase):
    """Nonce payload parser.

    The nonce payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                            Nonce Data                         ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar nonce_data: Random data generated by the transmitting entity
    """
    nonce_data: bytes = attr.ib(
        converter=bytes,
        validator=attr.validators.and_(
            attr.validators.instance_of(bytes),
            attr.validators.min_len(16),
            attr.validators.max_len(256)
        )
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.NONCE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        nonce_data_length = parser['payload_length'] - cls.HEADER_SIZE

        if nonce_data_length < 16:
            raise NotEnoughData(bytes_needed=16 - nonce_data_length)
        if nonce_data_length > 256:
            raise TooMuchData(bytes_needed=nonce_data_length - 256)

        parser.parse_raw('nonce_data', nonce_data_length)

        payload = cls(
            flags=parser['flags'],
            nonce_data=parser['nonce_data']
        )
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.nonce_data)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev2PayloadKeyExchange(Ikev2PayloadBase):
    """Key exchange payload parser.

    The key exchange payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |   Diffie-Hellman Group Num    |           RESERVED            |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                       Key Exchange Data                       ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar dh_group: Diffie-Hellman group number
    :ivar key_exchange_data: Diffie-Hellman public value
    """
    dh_group: Ikev2DiffieHellmanGroup = attr.ib(validator=attr.validators.instance_of(Ikev2DiffieHellmanGroup))
    key_exchange_data: typing.Union[bytes, bytearray] = attr.ib(
        validator=attr.validators.instance_of((bytes, bytearray))
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.KE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric_enum_coded('dh_group', Ikev2DiffieHellmanGroup)
        parser.parse_numeric('reserved2', 2)

        key_exchange_length = parser['payload_length'] - 8
        parser.parse_raw('key_exchange_data', key_exchange_length)

        payload = cls(
            flags=parser['flags'],
            dh_group=parser['dh_group'],
            key_exchange_data=parser['key_exchange_data']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric(self.dh_group.value.code, 2)
        composer_payload.compose_numeric(0, 2)  # reserved2
        composer_payload.compose_raw(self.key_exchange_data)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev2PayloadDelete(Ikev2PayloadBase):
    """Delete payload parser.

    The Delete payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Protocol ID   |   SPI Size    |          Num of SPIs          |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~               Security Parameter Index(es) (SPI)              ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar protocol_id: Protocol ID
    :ivar spi_size: Length in octets of the SPI
    :ivar num_spis: Number of SPIs contained in the payload
    :ivar spis: List of Security Parameter Indexes
    """
    protocol_id: Ikev2ProtocolId = attr.ib(validator=attr.validators.instance_of(Ikev2ProtocolId))
    spis: typing.List[int] = attr.ib(validator=attr.validators.deep_iterable(
        member_validator=attr.validators.instance_of(int),
    ))

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.DELETE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('protocol_id', Ikev2ProtocolId)
        parser.parse_numeric('spi_size', 1)
        parser.parse_numeric('num_spis', 2)

        parser.parse_numeric_array('spis', parser['num_spis'], 8)
        payload = cls(
            flags=parser['flags'],
            protocol_id=parser['protocol_id'],
            spis=parser['spis']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric_enum_coded(self.protocol_id)
        composer_payload.compose_numeric(len(self.spis) * 8, 1)
        composer_payload.compose_numeric(len(self.spis), 2)

        composer_payload.compose_numeric_array(self.spis, 8)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev2NotifyTypeFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return Ikev2NotifyType

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class Ikev2PayloadNotifyBase(Ikev2PayloadBase):
    """Notify payload parser.

    The notify payload has the following format:

    .. code-block:: text
                         1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        | Next Payload  |C|  RESERVED   |         Payload Length        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |  Protocol ID  |   SPI Size    |      Notify Message Type      |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        ~                Security Parameter Index (SPI)                 ~
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                                                               |
        ~                       Notification Data                       ~
        |                                                               |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar protocol_id: Protocol ID (1 byte)
    :ivar spi_size: Size of SPI in bytes (1 byte)
    :ivar notify_message_type: Type of notification message (2 bytes)
    :ivar spi: Security Parameter Index (variable length)
    :ivar data: Notification data (variable length)
    """
    protocol_id: Ikev2ProtocolId = attr.ib(validator=attr.validators.instance_of(Ikev2ProtocolId))
    type: Ikev2NotifyType = attr.ib(validator=attr.validators.instance_of(Ikev2NotifyType))
    spi: bytes = attr.ib(converter=bytes, validator=attr.validators.instance_of(bytes))

    @classmethod
    @abc.abstractmethod
    def _parse_type(cls, parser, name):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_data(cls, parser, notification_data_length):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_data(self, composer):
        raise NotImplementedError()

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.NOTIFY

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('protocol_id', Ikev2ProtocolId)
        parser.parse_numeric('spi_size', 1)
        cls._parse_type(parser, 'type')

        if parser['spi_size'] > 0:
            parser.parse_raw('spi', parser['spi_size'])
            spi = parser['spi']
        else:
            spi = bytes()

        del parser['spi_size']
        if 'spi' in parser:
            del parser['spi']

        notification_data_length = parser['payload_length'] - (cls.HEADER_SIZE + 4) - len(spi)

        cls._parse_data(parser, notification_data_length)

        next_payload = parser['next_payload']
        del parser['next_payload']
        del parser['payload_length']
        payload = cls(
            **parser,
            spi=spi,
        )
        payload.next_payload = next_payload

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()

        composer_payload.compose_numeric_enum_coded(self.protocol_id)
        composer_payload.compose_numeric(len(self.spi), 1)
        composer_payload.compose_numeric_enum_coded(self.type)

        if self.spi:
            composer_payload.compose_raw(self.spi)

        self._compose_data(composer_payload)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev2PayloadNotifyNoData(Ikev2PayloadNotifyBase):
    @classmethod
    def _parse_data(cls, parser, notification_data_length):
        pass

    def _compose_data(self, composer):
        pass

    @classmethod
    @abc.abstractmethod
    def _get_message_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_parsable(name, Ikev2NotifyTypeFactory)

        if parser[name] != cls._get_message_type():
            raise InvalidType()


class Ikev2PayloadNotifyAuthenticationFailed(Ikev2PayloadNotifyNoData):
    @classmethod
    def _get_message_type(cls):
        return Ikev2NotifyType.AUTHENTICATION_FAILED


@attr.s
class Ikev2PayloadNotifyUnparsed(Ikev2PayloadNotifyBase):
    data: typing.Union[bytes, bytearray] = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse_data(cls, parser, notification_data_length):
        parser.parse_raw('data', notification_data_length)

    def _compose_data(self, composer):
        composer.compose_raw(self.data)

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_numeric_enum_coded(name, Ikev2NotifyType)


@attr.s
class Ikev2PayloadNotifyParsedBase(Ikev2PayloadNotifyBase):
    @classmethod
    @abc.abstractmethod
    def _parse_data(cls, parser, notification_data_length):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_data(self, composer):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_message_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_numeric_enum_coded(name, Ikev2NotifyType)

        if parser[name] != cls._get_message_type():
            raise InvalidType()


@attr.s
class Ikev2NotifyPayloadInvalidKe(Ikev2PayloadNotifyParsedBase):
    """Invalid KE payload notification data parser."""
    dh_group: Ikev2DiffieHellmanGroup = attr.ib(validator=attr.validators.instance_of(Ikev2DiffieHellmanGroup))

    @classmethod
    def _get_message_type(cls):
        return Ikev2NotifyType.INVALID_KE_PAYLOAD

    @classmethod
    def _parse_data(cls, parser, notification_data_length):
        parser.parse_numeric_enum_coded('dh_group', Ikev2DiffieHellmanGroup)

    def _compose_data(self, composer):
        composer.compose_numeric_enum_coded(self.dh_group)


@attr.s
class Ikev2NotifyPayloadCookie(Ikev2PayloadNotifyParsedBase):
    """Invalid KE payload notification data parser."""
    cookie: typing.Union[bytes, bytearray] = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _get_message_type(cls):
        return Ikev2NotifyType.COOKIE

    @classmethod
    def _parse_data(cls, parser, notification_data_length):
        parser.parse_raw('cookie', notification_data_length)

    def _compose_data(self, composer):
        composer.compose_raw(self.cookie)


class Ikev2NotifyPayloadVariantBase(VariantParsable):
    @classmethod
    @abc.abstractmethod
    def get_parsed_notifies(cls):
        raise NotImplementedError()

    @classmethod
    def _get_variants(cls):
        variants = cls.get_parsed_notifies()

        variants.update([
            (notify_type, (Ikev2PayloadNotifyUnparsed, ))
            for notify_type in Ikev2NotifyType
            if notify_type not in variants
        ])

        return variants


class Ikev2NotifyPayloadVariantResponder(Ikev2NotifyPayloadVariantBase):
    @classmethod
    def get_parsed_notifies(cls):
        return collections.OrderedDict([
            (Ikev2NotifyType.COOKIE, [Ikev2NotifyPayloadCookie, ]),
            (Ikev2NotifyType.INVALID_KE_PAYLOAD, [Ikev2NotifyPayloadInvalidKe, ]),
        ])


@attr.s
class Ikev2PayloadCertificateRequest(Ikev2PayloadBase):
    """Certificate Request payload parser.

    The Certificate Request payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Cert Encoding |                                               |
       +-+-+-+-+-+-+-+-+                                               |
       ~                       Certificate Data                        ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar cert_encoding: Certificate encoding type
    :ivar certificate_data: Certificate data
    """
    cert_encoding: int = attr.ib(validator=attr.validators.instance_of(int))
    certificate_data: typing.Union[bytes, bytearray] = attr.ib(
        validator=attr.validators.instance_of((bytes, bytearray))
    )

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.CERTREQ

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_numeric('cert_encoding', 1)
        parser.parse_raw('certificate_data', parser['payload_length'] - cls.HEADER_SIZE - 1)

        payload = cls(
            flags=parser['flags'],
            cert_encoding=parser['cert_encoding'],
            certificate_data=parser['certificate_data']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_numeric(self.cert_encoding, 1)
        composer_payload.compose_raw(self.certificate_data)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


@attr.s
class Ikev2PayloadVendorId(Ikev2PayloadBase):
    """Vendor ID payload parser.

    The Vendor ID payload has the following format:

    .. code-block:: text
                        1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       | Next Payload  |C|  RESERVED   |         Payload Length        |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                                                               |
       ~                        Vendor ID (VID)                        ~
       |                                                               |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    :ivar vendor_id: Vendor ID data
    """
    vendor_id: typing.Union[bytes, bytearray] = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.VENDOR_ID

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        parser.parse_raw('vendor_id', parser['payload_length'] - cls.HEADER_SIZE)

        payload = cls(
            flags=parser['flags'],
            vendor_id=parser['vendor_id']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.vendor_id)

        composer_header = self.compose_header(composer_payload.composed_length)

        return composer_header.composed_bytes + composer_payload.composed_bytes


IKEV2_PAYLOAD_CLASSES_BY_TYPE = {
    Ikev2PayloadType.SA: Ikev2PayloadSecurityAssociation,
    Ikev2PayloadType.NONCE: Ikev2PayloadNonce,
    Ikev2PayloadType.KE: Ikev2PayloadKeyExchange,
    Ikev2PayloadType.NOTIFY: Ikev2NotifyPayloadVariantResponder,
    Ikev2PayloadType.CERTREQ: Ikev2PayloadCertificateRequest,
    Ikev2PayloadType.VENDOR_ID: Ikev2PayloadVendorId,
}


class Ikev2TransformVariantBase(VariantParsable):
    @classmethod
    @abc.abstractmethod
    def get_parsed_transforms(cls):
        raise NotImplementedError()

    @classmethod
    def _get_variants(cls):
        variants = cls.get_parsed_transforms()

        variants.update([
            (Ikev2TransformType.ENCR, [Ikev2TransformEncryptionAlgorithm, ]),
            (Ikev2TransformType.PRF, [Ikev2TransformPrf, ]),
            (Ikev2TransformType.DH, [Ikev2TransformDhGroup, ]),
            (Ikev2TransformType.INTEG, [Ikev2TransformIntegrity, ]),
        ])

        return variants


class Ikev2TransformVariantInitiator(Ikev2TransformVariantBase):
    @classmethod
    def get_parsed_transforms(cls):
        return collections.OrderedDict()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

from cryptodatahub.ike.algorithm import (
    Ikev1PayloadType,
    Ikev2PayloadType,
    Ikev2NotifyType,
    Ikev2ProtocolId,
    Ikev2TransformType,
    Ikev2PseudorandomFunction,
)
from cryptoparser.common.parse import ComposerBinary
from cryptoparser.ike.ikev1 import Ikev1PayloadBase, Ikev1PayloadDoiProtocolSpiBase
from cryptoparser.ike.ikev2 import (
    Ikev2PayloadBase,
    Ikev2PayloadNotifyBase,
    Ikev2PayloadNotifyNoData,
    Transform,
    TransformNextPayload,
)


class Ikev1PayloadBaseTest(Ikev1PayloadBase):
    """Concrete implementation of Ikev1PayloadBase for testing."""

    def __init__(self, test_data):
        super().__init__()
        self.test_data = test_data
        self.next_payload = Ikev1PayloadType.NONE

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.NONE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        test_data_length = parser['payload_length'] - cls.HEADER_SIZE
        parser.parse_raw('test_data', test_data_length)

        payload = cls(
            test_data=parser['test_data']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.test_data)

        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev1PayloadDoiProtocolSpiBaseTest(Ikev1PayloadDoiProtocolSpiBase):
    """Concrete implementation of Ikev1PayloadDoiProtocolSpiBase for testing."""

    def __init__(self, doi, protocol_id, spi_size, extra_data=b''):
        super().__init__(doi=doi, protocol_id=protocol_id, spi_size=spi_size)
        self.extra_data = extra_data
        self.next_payload = Ikev1PayloadType.NONE

    @classmethod
    def get_payload_type(cls):
        return Ikev1PayloadType.NONE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        extra_data_length = parser['payload_length'] - parser.parsed_length
        parser.parse_raw('extra_data', extra_data_length)

        payload = cls(
            doi=parser['doi'],
            protocol_id=parser['protocol_id'],
            spi_size=parser['spi_size'],
            extra_data=bytes(parser['extra_data'])
        )
        payload.next_payload = parser['next_payload']
        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        self._compose_doi_protocol_spi(composer_payload)
        composer_payload.compose_raw(self.extra_data)

        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev2PayloadBaseTest(Ikev2PayloadBase):
    """Concrete implementation of Ikev2PayloadBase for testing."""

    def __init__(self, flags, test_data):
        super().__init__(flags=flags)
        self.test_data = test_data
        self.next_payload = Ikev2PayloadType.NONE

    @classmethod
    def get_payload_type(cls):
        return Ikev2PayloadType.NONE

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)
        test_data_length = parser['payload_length'] - cls.HEADER_SIZE
        parser.parse_raw('test_data', test_data_length)

        payload = cls(
            flags=parser['flags'],
            test_data=parser['test_data']
        )
        payload.next_payload = parser['next_payload']

        return payload, parser.parsed_length

    def compose(self):
        composer_payload = ComposerBinary()
        composer_payload.compose_raw(self.test_data)

        composer_header = self.compose_header(composer_payload.composed_length)
        return composer_header.composed_bytes + composer_payload.composed_bytes


class Ikev2PayloadNotifyNoDataTest(Ikev2PayloadNotifyNoData):
    """Concrete implementation of Ikev2PayloadNotifyNoData for testing."""

    def __init__(self, flags, protocol_id, notify_type, spi):
        super().__init__(flags=flags, protocol_id=protocol_id, type=notify_type, spi=spi)
        self.next_payload = Ikev2PayloadType.NONE

    @classmethod
    def _get_message_type(cls):
        return Ikev2NotifyType.AUTHENTICATION_FAILED

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('protocol_id', Ikev2ProtocolId)
        parser.parse_numeric('spi_size', 1)
        cls._parse_type(parser, 'type')

        if parser['spi_size'] > 0:
            parser.parse_raw('spi', parser['spi_size'])
            spi = parser['spi']
        else:
            spi = bytes()

        del parser['spi_size']
        if 'spi' in parser:
            del parser['spi']

        notification_data_length = parser['payload_length'] - (cls.HEADER_SIZE + 4)

        cls._parse_data(parser, notification_data_length)

        next_payload = parser['next_payload']
        del parser['next_payload']
        del parser['payload_length']

        payload = cls(
            flags=parser['flags'],
            protocol_id=parser['protocol_id'],
            notify_type=parser['type'],
            spi=spi,
        )
        payload.next_payload = next_payload

        return payload, parser.parsed_length


class Ikev2PayloadNotifyBaseTest(Ikev2PayloadNotifyBase):
    """Concrete implementation of Ikev2PayloadNotifyBase for testing."""

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self, flags, protocol_id, notify_type, spi, test_data
    ):
        super().__init__(flags=flags, protocol_id=protocol_id, type=notify_type, spi=spi)
        self.test_data = test_data
        self.next_payload = Ikev2PayloadType.NONE

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_numeric_enum_coded(name, Ikev2NotifyType)

    @classmethod
    def _parse_data(cls, parser, notification_data_length):
        if notification_data_length > 0:
            parser.parse_raw('test_data', notification_data_length)
        else:
            # For empty data, we need to handle it differently since parser is not a dict
            pass

    def _compose_data(self, composer):
        composer.compose_raw(self.test_data)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric_enum_coded('protocol_id', Ikev2ProtocolId)
        parser.parse_numeric('spi_size', 1)
        cls._parse_type(parser, 'type')

        if parser['spi_size'] > 0:
            parser.parse_raw('spi', parser['spi_size'])
            spi = parser['spi']
        else:
            spi = bytes()

        del parser['spi_size']
        if 'spi' in parser:
            del parser['spi']

        notification_data_length = parser['payload_length'] - (cls.HEADER_SIZE + 4) - len(spi)

        cls._parse_data(parser, notification_data_length)

        next_payload = parser['next_payload']
        del parser['next_payload']
        del parser['payload_length']

        test_data = parser.get('test_data', b'')
        if 'test_data' in parser:
            del parser['test_data']

        payload = cls(
            flags=parser['flags'],
            protocol_id=parser['protocol_id'],
            notify_type=parser['type'],
            spi=spi,
            test_data=test_data,
        )
        payload.next_payload = next_payload

        return payload, parser.parsed_length


class TransformTest(Transform):
    """Concrete implementation of Transform base class for testing."""

    def __init__(self, transform_id):
        super().__init__(transform_id=transform_id)
        self.next_payload = TransformNextPayload.LAST

    @classmethod
    def get_transform_type(cls):
        return Ikev2TransformType.PRF

    @classmethod
    def _get_transform_id_class(cls):
        return Ikev2PseudorandomFunction

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        transform = cls(
            transform_id=parser['transform_id'],
        )
        transform.next_payload = parser['next_payload']

        return transform, parser.parsed_length

    def compose(self):
        return self.compose_header(transform_length=0).composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import collections
import enum
import attr

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.tls.algorithm import (
    TlsCertificateCompressionAlgorithm,
    TlsExtensionType,
    TlsNamedCurve,
    TlsNextProtocolName,
    TlsProtocolName,
    TlsPskKeyExchangeMode,
    TlsTokenBindingParamater,
)

from cryptoparser.common.base import (
    OneByteEnumParsable,
    Opaque,
    OpaqueParam,
    ProtocolVersionMajorMinorBase,
    OpaqueEnumParsable,
    TwoByteEnumParsable,
    VariantParsable,
    Vector,
    VectorEnumCodeNumeric,
    VectorEnumCodeString,
    VectorParamEnumCodeNumeric,
    VectorParamEnumCodeString,
    VectorParamNumeric,
    VectorParamParsable,
    VectorParsable,
    VectorParsableDerived,
)
from cryptoparser.common.exception import NotEnoughData, InvalidType
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.x509 import SignedCertificateTimestampList
from cryptoparser.tls.algorithm import (
    TlsECPointFormatFactory,
    TlsNamedCurveFactory,
    TlsSignatureAndHashAlgorithmFactory,
)
from cryptoparser.tls.grease import TlsInvalidTypeOneByte, TlsInvalidTypeTwoByte
from cryptoparser.tls.version import TlsProtocolVersion


class TlsExtensionTypeFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsExtensionType

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsExtensionsBase(VectorParsable):
    @classmethod
    @abc.abstractmethod
    def get_param(cls):
        raise NotImplementedError()

    def get_item_by_type(self, extension_type):
        try:
            item = next(
                extension
                for extension in self
                if extension.extension_type == extension_type
            )
        except StopIteration as e:
            raise KeyError from e

        return item


class TlsExtensionsClient(TlsExtensionsBase):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsExtensionVariantClient,
            fallback_class=TlsExtensionUnparsed,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


class TlsExtensionsServer(TlsExtensionsBase):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsExtensionVariantServer,
            fallback_class=TlsExtensionUnparsed,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


@attr.s
class TlsExtensionBase(ParsableBase):
    extension_type = attr.ib(init=False, validator=attr.validators.instance_of(TlsExtensionType))

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_type(cls, parser, name):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_type(self, composer):
        raise NotImplementedError()

    @classmethod
    def _check_header(cls, parsable):
        parser = ParserBinary(parsable)

        cls._parse_type(parser, 'extension_type')
        parser.parse_numeric('extension_length', 2)

        if parser.unparsed_length < parser['extension_length']:
            raise NotEnoughData(parser['extension_length'] + parser.parsed_length)

        return parser

    def _compose_header(self, payload_length):
        header_composer = ComposerBinary()

        self._compose_type(header_composer)
        header_composer.compose_numeric(payload_length, 2)

        return header_composer.composed_bytes


@attr.s
class TlsExtensionUnparsed(TlsExtensionBase):
    extension_type = attr.ib(validator=attr.validators.instance_of((TlsExtensionType, TlsInvalidTypeTwoByte)))
    extension_data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_parsable(name, TlsInvalidTypeTwoByte)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._check_header(parsable)

        parser.parse_raw('extension_data', parser['extension_length'])

        return TlsExtensionUnparsed(parser['extension_type'], parser['extension_data']), parser.parsed_length

    def _compose_type(self, composer):
        composer.compose_parsable(self.extension_type)

    def compose(self):
        return self._compose_header(len(self.extension_data)) + self.extension_data


@attr.s
class TlsExtensionParsed(TlsExtensionBase):
    def __attrs_post_init__(self):
        self.extension_type = self.get_extension_type()

        attr.validate(self)

    @classmethod
    def _parse_type(cls, parser, name):
        parser.parse_parsable(name, TlsExtensionTypeFactory)

    def _compose_type(self, composer):
        composer.compose_numeric_enum_coded(self.extension_type)

    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        parser = super(TlsExtensionParsed, cls)._check_header(parsable)

        if parser['extension_type'] != cls.get_extension_type():
            raise InvalidType()

        return parser


class TlsExtensionUnusedData(TlsExtensionParsed):
    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_raw('extension_data', parser['extension_length'])

        if parser['extension_data']:
            raise InvalidValue(parser['extension_data'], cls)

        return cls(), parser.parsed_length

    def compose(self):
        return self._compose_header(0)

    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()


class TlsServerNameType(enum.IntEnum):
    HOST_NAME = 0x00


class TlsServerName(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(
            item_size=1,
            min_byte_num=1,
            max_byte_num=2 ** 16 - 1,
        )


@attr.s
class TlsExtensionServerNameClient(TlsExtensionParsed):
    host_name = attr.ib(validator=attr.validators.instance_of(str))
    name_type = attr.ib(validator=attr.validators.in_(TlsServerNameType), default=TlsServerNameType.HOST_NAME)

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SERVER_NAME

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionServerNameClient, cls)._parse_header(parsable)

        parser.parse_numeric('server_name_list_length', 2)
        parser.parse_numeric('server_name_type', 1, TlsServerNameType)
        parser.parse_parsable('server_name', TlsServerName)

        return cls(bytearray(parser['server_name']).decode('idna')), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        idna_encoded_host_name = self.host_name.encode('idna')

        composer.compose_numeric(3 + len(idna_encoded_host_name), 2)
        composer.compose_numeric(self.name_type, 1)

        composer.compose_bytes(idna_encoded_host_name, 2)

        header_bytes = self._compose_header(composer.composed_length)

        return header_bytes + composer.composed_bytes


@attr.s
class TlsExtensionServerNameServer(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SERVER_NAME


class TlsECPointFormatVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsECPointFormatFactory,
            fallback_class=TlsInvalidTypeOneByte,
            min_byte_num=1,
            max_byte_num=2 ** 8 - 1,
        )


@attr.s
class TlsExtensionECPointFormats(TlsExtensionParsed):
    point_formats = attr.ib(
        converter=TlsECPointFormatVector,
        validator=attr.validators.instance_of(TlsECPointFormatVector)
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.EC_POINT_FORMATS

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionECPointFormats, cls)._parse_header(parsable)

        parser.parse_parsable('point_formats', TlsECPointFormatVector)

        return TlsExtensionECPointFormats(parser['point_formats']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.point_formats)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsEllipticCurveVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsNamedCurveFactory,
            fallback_class=TlsInvalidTypeTwoByte,
            min_byte_num=1, max_byte_num=2 ** 16 - 1
        )


@attr.s
class TlsExtensionEllipticCurves(TlsExtensionParsed):
    elliptic_curves = attr.ib(
        converter=TlsEllipticCurveVector,
        validator=attr.validators.instance_of(TlsEllipticCurveVector)
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SUPPORTED_GROUPS

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionEllipticCurves, cls)._parse_header(parsable)

        parser.parse_parsable('elliptic_curves', TlsEllipticCurveVector)

        return TlsExtensionEllipticCurves(parser['elliptic_curves']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.elliptic_curves)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsSupportedVersionVector(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsProtocolVersion,
            fallback_class=TlsInvalidTypeTwoByte,
            min_byte_num=2, max_byte_num=2 ** 8 - 2
        )


@attr.s
class TlsExtensionSupportedVersionsBase(TlsExtensionParsed):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SUPPORTED_VERSIONS

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class TlsExtensionSupportedVersionsClient(TlsExtensionSupportedVersionsBase):
    supported_versions = attr.ib(
        converter=TlsSupportedVersionVector,
        validator=attr.validators.instance_of(TlsSupportedVersionVector)
    )

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionSupportedVersionsClient, cls)._parse_header(parsable)

        parser.parse_parsable('supported_versions', TlsSupportedVersionVector)

        return TlsExtensionSupportedVersionsClient(parser['supported_versions']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.supported_versions)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionSupportedVersionsServer(TlsExtensionSupportedVersionsBase):
    selected_version = attr.ib(validator=attr.validators.instance_of(TlsProtocolVersion))

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionSupportedVersionsServer, cls)._parse_header(parsable)

        parser.parse_parsable('selected_version', TlsProtocolVersion)

        return TlsExtensionSupportedVersionsServer(parser['selected_version']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.selected_version)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsSignatureAndHashAlgorithmVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsSignatureAndHashAlgorithmFactory,
            fallback_class=TlsInvalidTypeTwoByte,
            min_byte_num=2, max_byte_num=2 ** 16 - 2
        )


@attr.s
class TlsExtensionSignatureAlgorithmsBase(TlsExtensionParsed):
    hash_and_signature_algorithms = attr.ib(
        converter=TlsSignatureAndHashAlgorithmVector,
        validator=attr.validators.instance_of(TlsSignatureAndHashAlgorithmVector)
    )

    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionSignatureAlgorithmsBase, cls)._parse_header(parsable)

        parser.parse_parsable('hash_and_signature_algorithms', TlsSignatureAndHashAlgorithmVector)

        return cls(parser['hash_and_signature_algorithms']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.hash_and_signature_algorithms)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsExtensionSignatureAlgorithms(TlsExtensionSignatureAlgorithmsBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SIGNATURE_ALGORITHMS


class TlsExtensionSignatureAlgorithmsCert(TlsExtensionSignatureAlgorithmsBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SIGNATURE_ALGORITHMS_CERT


class TlsExtensionDelegatedCredentials(TlsExtensionSignatureAlgorithmsBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.DELEGATED_CREDENTIALS


class TlsKeyExchangeVector(Vector):
    @classmethod
    def get_param(cls):
        return VectorParamNumeric(item_size=1, min_byte_num=1, max_byte_num=2 ** 16 - 1)


@attr.s
class TlsKeyShareEntry(ParsableBase):
    group = attr.ib(validator=attr.validators.instance_of(TlsNamedCurve))
    key_exchange = attr.ib(converter=TlsKeyExchangeVector)

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('group', TlsNamedCurveFactory)
        parser.parse_parsable('key_exchange', TlsKeyExchangeVector)

        return TlsKeyShareEntry(parser['group'], parser['key_exchange']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric_enum_coded(self.group)
        composer.compose_parsable(self.key_exchange)

        return composer.composed_bytes


@attr.s
class TlsKeyShareEntryInvalidType(ParsableBase):
    group = attr.ib(validator=attr.validators.instance_of(TlsInvalidTypeTwoByte))
    data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_parsable('group', TlsInvalidTypeTwoByte)
        parser.parse_bytes('data', 2)

        return TlsKeyShareEntryInvalidType(parser['group'], parser['data']), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_parsable(self.group)
        composer.compose_bytes(self.data, 2)

        return composer.composed_bytes


class TlsKeyShareEntryVector(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsKeyShareEntry,
            fallback_class=TlsKeyShareEntryInvalidType,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


class TlsExtensionKeyShareBase(TlsExtensionParsed):
    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class TlsExtensionKeyShareServer(TlsExtensionKeyShareBase):
    key_share_entry = attr.ib(validator=attr.validators.instance_of(TlsKeyShareEntry))

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.KEY_SHARE

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionKeyShareServer, cls)._parse_header(parsable)

        parser.parse_parsable('key_share_entry', TlsKeyShareEntry)

        return TlsExtensionKeyShareServer(parser['key_share_entry']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.key_share_entry)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionKeyShareClientHelloRetry(TlsExtensionKeyShareBase):
    selected_group = attr.ib(converter=TlsNamedCurve)

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.KEY_SHARE

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionKeyShareClientHelloRetry, cls)._parse_header(parsable)

        if parser['extension_length'] != 2:
            raise InvalidType()

        parser.parse_parsable('selected_group', TlsNamedCurveFactory)

        return cls(parser['selected_group']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_numeric_enum_coded(self.selected_group)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionKeyShareEntriesClientBase(TlsExtensionKeyShareBase):
    key_share_entries = attr.ib(converter=TlsKeyShareEntryVector)

    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionKeyShareEntriesClientBase, cls)._parse_header(parsable)

        parser.parse_parsable('key_share_entries', TlsKeyShareEntryVector)

        return cls(parser['key_share_entries']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.key_share_entries)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsExtensionKeyShareClient(TlsExtensionKeyShareEntriesClientBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.KEY_SHARE


class TlsExtensionKeyShareReservedClient(TlsExtensionKeyShareEntriesClientBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.KEY_SHARE_RESERVED


class TlsCertificateStatusType(enum.IntEnum):
    OCSP = 1


class TlsCertificateStatusRequestExtensions(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=0,
            max_byte_num=2 ** 16 - 1,
        )


class TlsCertificateStatusRequestResponderId(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=1,
            max_byte_num=2 ** 16 - 1,
        )


class TlsCertificateStatusRequestResponderIdList(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=TlsCertificateStatusRequestResponderId,
            fallback_class=None,
            min_byte_num=0, max_byte_num=2 ** 16 - 1
        )


class TlsExtensionCertificateStatusRequestClient(TlsExtensionParsed):
    def __init__(self, responder_id_list=(), extensions=()):
        super().__init__()

        self.responder_id_list = TlsCertificateStatusRequestResponderIdList(responder_id_list)
        self.request_extensions = TlsCertificateStatusRequestExtensions(extensions)

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.STATUS_REQUEST

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionCertificateStatusRequestClient, cls)._parse_header(parsable)

        parser.parse_numeric('status_type', 1, TlsCertificateStatusType)
        parser.parse_parsable('responder_id_list', TlsCertificateStatusRequestResponderIdList)
        parser.parse_parsable('extensions', TlsCertificateStatusRequestExtensions)

        return cls(
            parser['responder_id_list'],
            parser['extensions'],
        ), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_numeric(TlsCertificateStatusType.OCSP, 1)
        payload_composer.compose_parsable(self.responder_id_list)
        payload_composer.compose_parsable(self.request_extensions)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsExtensionCertificateStatusRequestServer(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.STATUS_REQUEST


class TlsRenegotiatedConnection(Opaque):
    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=0,
            max_byte_num=2 ** 8 - 1,
        )


@attr.s
class TlsExtensionRenegotiationInfo(TlsExtensionParsed):
    renegotiated_connection = attr.ib(
        default=TlsRenegotiatedConnection([]),
        validator=attr.validators.instance_of(TlsRenegotiatedConnection)
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.RENEGOTIATION_INFO

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionRenegotiationInfo, cls)._parse_header(parsable)

        parser.parse_parsable('renegotiated_connection', TlsRenegotiatedConnection)

        return TlsExtensionRenegotiationInfo(parser['renegotiated_connection']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.renegotiated_connection)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionSessionTicket(TlsExtensionParsed):
    session_ticket = attr.ib(
        default=bytearray([]),
        validator=attr.validators.instance_of((bytes, bytearray))
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SESSION_TICKET

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionSessionTicket, cls)._parse_header(parsable)

        parser.parse_raw('session_ticket', parser['extension_length'])

        return TlsExtensionSessionTicket(parser['session_ticket']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_raw(self.session_ticket)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsProtocolNameFactory(OpaqueEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsProtocolName

    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=1, max_byte_num=2 ** 8 - 1
        )


class TlsProtocolNameList(VectorEnumCodeString):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeString(
            item_class=TlsProtocolNameFactory,
            min_byte_num=2, max_byte_num=2 ** 16 - 1
        )


@attr.s
class TlsExtensionApplicationLayerProtocolBase(TlsExtensionParsed):
    protocol_names = attr.ib(
        converter=TlsProtocolNameList,
        validator=attr.validators.instance_of(TlsProtocolNameList),
    )

    @classmethod
    @abc.abstractmethod
    def get_extension_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionApplicationLayerProtocolBase, cls)._parse_header(parsable)

        parser.parse_parsable('protocol_names', TlsProtocolNameList)

        return cls(parser['protocol_names']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.protocol_names)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsExtensionApplicationLayerProtocolNegotiation(TlsExtensionApplicationLayerProtocolBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION


class TlsExtensionApplicationLayerProtocolSettings(TlsExtensionApplicationLayerProtocolBase):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.APPLICATION_LAYER_PROTOCOL_SETTINGS


class TlsExtensionNextProtocolNegotiationClient(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION


class TlsNextProtocolNameFactory(OpaqueEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsNextProtocolName

    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=1, max_byte_num=2 ** 8 - 1
        )


class TlsNextProtocolNameList(VectorEnumCodeString):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeString(
            item_class=TlsNextProtocolNameFactory,
            min_byte_num=1, max_byte_num=2 ** 16 - 1
        )


@attr.s
class TlsExtensionNextProtocolNegotiationServer(TlsExtensionParsed):
    protocol_names = attr.ib(
        converter=TlsNextProtocolNameList,
        validator=attr.validators.instance_of(TlsNextProtocolNameList),
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        cls._parse_type(parser, 'extension_type')
        if parser['extension_type'] != cls.get_extension_type():
            raise InvalidType()

        parser.parse_parsable('protocol_names', TlsNextProtocolNameList)

        return cls(parser['protocol_names']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()
        payload_composer.compose_parsable(self.protocol_names)

        header_composer = ComposerBinary()
        self._compose_type(header_composer)

        return header_composer.composed_bytes + payload_composer.composed_bytes


class TlsExtensionChannelId(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.CHANNEL_ID


class TlsExtensionEncryptThenMAC(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.ENCRYPT_THEN_MAC


class TlsExtensionExtendedMasterSecret(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.EXTENDED_MASTER_SECRET


@attr.s
class TlsExtensionShortRecordHeader(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SHORT_RECORD_HEADER


class TlsTokenBindingProtocolVersion(ProtocolVersionMajorMinorBase):
    pass


class TlsTokenBindingParamaterFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsTokenBindingParamater

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsTokenBindingParamaterVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsTokenBindingParamaterFactory,
            fallback_class=TlsInvalidTypeOneByte,
            min_byte_num=1,
            max_byte_num=2 ** 8 - 1,
        )


@attr.s
class TlsExtensionTokenBinding(TlsExtensionParsed):
    protocol_version = attr.ib(validator=attr.validators.instance_of(TlsTokenBindingProtocolVersion))
    parameters = attr.ib(
        converter=TlsTokenBindingParamaterVector,
        validator=attr.validators.instance_of(TlsTokenBindingParamaterVector)
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.TOKEN_BINDING

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionTokenBinding, cls)._parse_header(parsable)

        parser.parse_parsable('protocol_version', TlsTokenBindingProtocolVersion)
        parser.parse_parsable('parameters', TlsTokenBindingParamaterVector)

        return cls(parser['protocol_version'], parser['parameters']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.protocol_version)
        payload_composer.compose_parsable(self.parameters)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsPskKeyExchangeModeFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsPskKeyExchangeMode

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsPskKeyExchangeModeVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsPskKeyExchangeModeFactory,
            fallback_class=TlsInvalidTypeOneByte,
            min_byte_num=1,
            max_byte_num=2 ** 8 - 1,
        )


@attr.s
class TlsExtensionPskKeyExchangeModes(TlsExtensionParsed):
    key_exchange_modes = attr.ib(
        converter=TlsPskKeyExchangeModeVector,
        validator=attr.validators.instance_of(TlsPskKeyExchangeModeVector)
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.PSK_KEY_EXCHANGE_MODES

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionPskKeyExchangeModes, cls)._parse_header(parsable)

        parser.parse_parsable('key_exchange_modes', TlsPskKeyExchangeModeVector)

        return TlsExtensionPskKeyExchangeModes(parser['key_exchange_modes']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.key_exchange_modes)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionRecordSizeLimit(TlsExtensionParsed):
    record_size_limit = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.RECORD_SIZE_LIMIT

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_numeric('record_size_limit', 2)

        return TlsExtensionRecordSizeLimit(parser['record_size_limit']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_numeric(self.record_size_limit, 2)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionSignedCertificateTimestampClient(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP


@attr.s
class TlsExtensionSignedCertificateTimestampServer(TlsExtensionParsed):
    scts = attr.ib(
        converter=SignedCertificateTimestampList,
        validator=attr.validators.optional(attr.validators.instance_of(SignedCertificateTimestampList))
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP

    @classmethod
    def _parse(cls, parsable):
        parser = super(cls, cls)._parse_header(parsable)

        parser.parse_parsable('scts', SignedCertificateTimestampList)

        return cls(parser['scts']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.scts)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsCertificateCompressionAlgorithmFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsCertificateCompressionAlgorithm

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsCertificateCompressionAlgorithmVector(VectorEnumCodeNumeric):
    @classmethod
    def get_param(cls):
        return VectorParamEnumCodeNumeric(
            item_class=TlsCertificateCompressionAlgorithmFactory,
            fallback_class=TlsInvalidTypeTwoByte,
            min_byte_num=2, max_byte_num=2 ** 8 - 2
        )


@attr.s
class TlsExtensionCompressCertificate(TlsExtensionParsed):
    compression_algorithms = attr.ib(
        converter=TlsCertificateCompressionAlgorithmVector,
        validator=attr.validators.instance_of(TlsCertificateCompressionAlgorithmVector),
    )

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.COMPRESS_CERTIFICATE

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionCompressCertificate, cls)._parse_header(parsable)

        parser.parse_parsable('compression_algorithms', TlsCertificateCompressionAlgorithmVector)

        return cls(parser['compression_algorithms']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_parsable(self.compression_algorithms)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionPadding(TlsExtensionParsed):
    length = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.PADDING

    @classmethod
    def _parse(cls, parsable):
        parser = super(TlsExtensionPadding, cls)._parse_header(parsable)

        parser.parse_raw('padding', parser['extension_length'])
        try:
            non_zero_int = next(byte for byte in parser['padding'] if byte != 0)
            raise InvalidValue(bytes((non_zero_int,)), cls)
        except StopIteration:
            pass

        return cls(parser['extension_length']), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_raw(self.length * b'\x00')

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsEncryptedClientHelloType(enum.IntEnum):
    OUTER = 0
    INNER = 1


@attr.s
class TlsExtensionEncryptedClientHelloBase(TlsExtensionParsed):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.ENCRYPTED_CLIENT_HELLO

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_encrypted_client_hello_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        parser = super(TlsExtensionEncryptedClientHelloBase, cls)._parse_header(parsable)

        parser.parse_numeric('client_hello_type', 1, TlsEncryptedClientHelloType)
        if parser['client_hello_type'] != cls.get_encrypted_client_hello_type():
            raise InvalidType()

        return parser

    def compose_type(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.get_encrypted_client_hello_type(), 1)

        return composer


@attr.s
class TlsExtensionEncryptedClientHelloInner(TlsExtensionEncryptedClientHelloBase):
    @classmethod
    def get_encrypted_client_hello_type(cls):
        return TlsEncryptedClientHelloType.INNER

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        return cls(), parser.parsed_length

    def compose(self):
        payload_composer = self.compose_type()

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


@attr.s
class TlsExtensionEncryptedClientHelloOuter(TlsExtensionEncryptedClientHelloBase):
    data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def get_encrypted_client_hello_type(cls):
        return TlsEncryptedClientHelloType.OUTER

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_raw('data', parser.unparsed_length)

        return cls(parser['data']), parser.parsed_length

    def compose(self):
        payload_composer = self.compose_type()

        payload_composer.compose_raw(self.data)

        header_bytes = self._compose_header(payload_composer.composed_length)

        return header_bytes + payload_composer.composed_bytes


class TlsExtensionPostHandshakeAuthentication(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.POST_HANDSHAKE_AUTH


class TlsExtensionVariantBase(VariantParsable):
    @classmethod
    @abc.abstractmethod
    def get_parsed_extensions(cls):
        raise NotImplementedError()

    @classmethod
    def _get_variants(cls):
        variants = cls.get_parsed_extensions()

        variants.update([
            (extension_type, (TlsExtensionUnparsed, ))
            for extension_type in TlsExtensionType
            if extension_type not in variants
        ])

        return variants


class TlsExtensionVariantClient(TlsExtensionVariantBase):
    @classmethod
    def get_parsed_extensions(cls):
        return collections.OrderedDict([
            (TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION,
                [TlsExtensionApplicationLayerProtocolNegotiation, ]),
            (TlsExtensionType.APPLICATION_LAYER_PROTOCOL_SETTINGS,
                [TlsExtensionApplicationLayerProtocolSettings, ]),
            (TlsExtensionType.CHANNEL_ID, [TlsExtensionChannelId, ]),
            (TlsExtensionType.COMPRESS_CERTIFICATE, [TlsExtensionCompressCertificate, ]),
            (TlsExtensionType.ENCRYPT_THEN_MAC, [TlsExtensionEncryptThenMAC, ]),
            (TlsExtensionType.EXTENDED_MASTER_SECRET, [TlsExtensionExtendedMasterSecret, ]),
            (TlsExtensionType.RENEGOTIATION_INFO, [TlsExtensionRenegotiationInfo, ]),
            (TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION, [TlsExtensionNextProtocolNegotiationClient, ]),
            (TlsExtensionType.PADDING, [TlsExtensionPadding, ]),
            (TlsExtensionType.SERVER_NAME, [TlsExtensionServerNameClient, ]),
            (TlsExtensionType.SESSION_TICKET, [TlsExtensionSessionTicket, ]),
            (TlsExtensionType.STATUS_REQUEST, [TlsExtensionCertificateStatusRequestClient, ]),
            (TlsExtensionType.SUPPORTED_GROUPS, [TlsExtensionEllipticCurves, ]),
            (TlsExtensionType.DELEGATED_CREDENTIALS, [TlsExtensionDelegatedCredentials, ]),
            (TlsExtensionType.EC_POINT_FORMATS, [TlsExtensionECPointFormats, ]),
            (TlsExtensionType.KEY_SHARE, [TlsExtensionKeyShareClient, ]),
            (TlsExtensionType.KEY_SHARE_RESERVED, [TlsExtensionKeyShareReservedClient, ]),
            (TlsExtensionType.POST_HANDSHAKE_AUTH, [TlsExtensionPostHandshakeAuthentication, ]),
            (TlsExtensionType.PSK_KEY_EXCHANGE_MODES, [TlsExtensionPskKeyExchangeModes, ]),
            (TlsExtensionType.RECORD_SIZE_LIMIT, [TlsExtensionRecordSizeLimit, ]),
            (TlsExtensionType.SHORT_RECORD_HEADER, [TlsExtensionShortRecordHeader, ]),
            (TlsExtensionType.SIGNATURE_ALGORITHMS, [TlsExtensionSignatureAlgorithms, ]),
            (TlsExtensionType.SIGNATURE_ALGORITHMS_CERT, [TlsExtensionSignatureAlgorithmsCert, ]),
            (TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP, [TlsExtensionSignedCertificateTimestampClient, ]),
            (TlsExtensionType.SUPPORTED_VERSIONS, [TlsExtensionSupportedVersionsClient, ]),
            (TlsExtensionType.ENCRYPTED_CLIENT_HELLO, [TlsExtensionEncryptedClientHelloInner,
                                                       TlsExtensionEncryptedClientHelloOuter]),
            (TlsExtensionType.TOKEN_BINDING, [TlsExtensionTokenBinding, ]),
        ])


class TlsExtensionVariantServer(TlsExtensionVariantBase):
    @classmethod
    def get_parsed_extensions(cls):
        return collections.OrderedDict([
            (TlsExtensionType.APPLICATION_LAYER_PROTOCOL_NEGOTIATION,
                [TlsExtensionApplicationLayerProtocolNegotiation, ]),
            (TlsExtensionType.CHANNEL_ID, [TlsExtensionChannelId, ]),
            (TlsExtensionType.EC_POINT_FORMATS, [TlsExtensionECPointFormats, ]),
            (TlsExtensionType.ENCRYPT_THEN_MAC, [TlsExtensionEncryptThenMAC, ]),
            (TlsExtensionType.EXTENDED_MASTER_SECRET, [TlsExtensionExtendedMasterSecret, ]),
            (TlsExtensionType.KEY_SHARE, [TlsExtensionKeyShareClientHelloRetry, TlsExtensionKeyShareServer]),
            (TlsExtensionType.NEXT_PROTOCOL_NEGOTIATION, [TlsExtensionNextProtocolNegotiationServer, ]),
            (TlsExtensionType.RECORD_SIZE_LIMIT, [TlsExtensionRecordSizeLimit, ]),
            (TlsExtensionType.RENEGOTIATION_INFO, [TlsExtensionRenegotiationInfo, ]),
            (TlsExtensionType.SERVER_NAME, [TlsExtensionServerNameServer, ]),
            (TlsExtensionType.SESSION_TICKET, [TlsExtensionSessionTicket, ]),
            (TlsExtensionType.SIGNED_CERTIFICATE_TIMESTAMP, [TlsExtensionSignedCertificateTimestampServer, ]),
            (TlsExtensionType.STATUS_REQUEST, [TlsExtensionCertificateStatusRequestServer, ]),
            (TlsExtensionType.SUPPORTED_VERSIONS, [TlsExtensionSupportedVersionsServer, ]),
        ])



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest
import collections

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.tls.openvpn import (
    OpenVpnPacketAckV1,
    OpenVpnPacketBase,
    OpenVpnPacketControlV1,
    OpenVpnPacketHardResetClientV2,
    OpenVpnPacketHardResetServerV2,
    OpenVpnPacketVariant,
    OpenVpnPacketWrapperTcp,
)


class TestOpenVpnPacketWrapperTcp(unittest.TestCase):
    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            OpenVpnPacketWrapperTcp.parse_exact_size(b'\x00\x02\x00')

        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        packet = OpenVpnPacketWrapperTcp.parse_exact_size(b'\x00\x08\x00\x01\x02\x03\x04\x05\x06\x07')
        self.assertEqual(packet.payload, b'\x00\x01\x02\x03\x04\x05\x06\x07')

    def test_compose(self):
        packet = OpenVpnPacketWrapperTcp(b'\x00\x01\x02\x03\x04\x05\x06\x07')
        self.assertEqual(packet.compose(), b'\x00\x08\x00\x01\x02\x03\x04\x05\x06\x07')


class TestOpenVpnPacketBase(unittest.TestCase):
    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            OpenVpnPacketBase.parse_header(b'\x00')

        self.assertEqual(context_manager.exception.bytes_needed, OpenVpnPacketBase.HEADER_SIZE - 1)

    def test_error_wrong_packet_type(self):
        with self.assertRaises(InvalidType):
            OpenVpnPacketAckV1.parse_header(b'\x00' * OpenVpnPacketBase.HEADER_SIZE)

    def test_packet_id_array(self):
        header_without_packet_array_dict = collections.OrderedDict([
            ('opcode', b'\x20'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
        ])
        header_without_packet_array_bytes = b''.join(header_without_packet_array_dict.values())

        session_id, packet_id_array, remote_session_id, header_length = OpenVpnPacketControlV1.parse_header(
            header_without_packet_array_bytes + b'\x00'
        )
        self.assertEqual(session_id, 0x0001020304050607)
        self.assertEqual(packet_id_array, [])
        self.assertEqual(remote_session_id, None)
        self.assertEqual(header_length, len(header_without_packet_array_bytes) + 1)

        packet_array_dict = collections.OrderedDict([
            ('packet_id_array_length', b'\x01'),
            ('packet_id_array', b'\x04\x05\x06\x07'),
            ('remote_session_id', b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        packet_array_bytes = b''.join(packet_array_dict.values())

        session_id, packet_id_array, remote_session_id, header_length = OpenVpnPacketControlV1.parse_header(
            header_without_packet_array_bytes + packet_array_bytes
        )
        self.assertEqual(session_id, 0x0001020304050607)
        self.assertEqual(packet_id_array, [0x04050607])
        self.assertEqual(remote_session_id, 0x08090a0b0c0d0e0f)
        self.assertEqual(header_length, len(header_without_packet_array_bytes) + len(packet_array_bytes))


class TestOpenVpnPacketControlV1(unittest.TestCase):
    def setUp(self):
        self.control_dict = collections.OrderedDict([
            ('opcode', b'\x20'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('packet_id_array_length', b'\x01'),
            ('packet_id_array', b'\x04\x05\x06\x07'),
            ('remote_session_id', b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
            ('packet_id', b'\x00\x01\x02\x03'),
            ('payload', b'\x00\x01\x02\x03'),
        ])
        self.control_bytes = b''.join(self.control_dict.values())

        self.control = OpenVpnPacketControlV1(
            session_id=0x0001020304050607,
            packet_id_array=[0x04050607],
            remote_session_id=0x08090a0b0c0d0e0f,
            packet_id=0x00010203,
            payload=b'\x00\x01\x02\x03',
        )

    def test_parse(self):
        control = OpenVpnPacketVariant.parse_exact_size(self.control_bytes)
        self.assertEqual(control.session_id, 0x0001020304050607)
        self.assertEqual(control.packet_id_array, [0x04050607])
        self.assertEqual(control.remote_session_id, 0x08090a0b0c0d0e0f)
        self.assertEqual(control.packet_id, 0x00010203)

    def test_compose(self):
        self.assertEqual(self.control.compose(), self.control_bytes)


class TestOpenVpnPacketAckV1(unittest.TestCase):
    def setUp(self):
        self.ack_dict = collections.OrderedDict([
            ('opcode', b'\x28'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('packet_id_array_length', b'\x01'),
            ('packet_id_array', b'\x04\x05\x06\x07'),
            ('remote_session_id', b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        self.ack_bytes = b''.join(self.ack_dict.values())

        self.ack = OpenVpnPacketAckV1(
            session_id=0x0001020304050607,
            packet_id_array=[0x04050607],
            remote_session_id=0x08090a0b0c0d0e0f,
        )

    def test_parse(self):
        ack = OpenVpnPacketVariant.parse_exact_size(self.ack_bytes)
        self.assertEqual(ack.session_id, 0x0001020304050607)
        self.assertEqual(ack.packet_id_array, [0x04050607])

    def test_compose(self):
        self.assertEqual(self.ack.compose(), self.ack_bytes)


class TestOpenVpnPacketHardResetClientV2(unittest.TestCase):
    def setUp(self):
        self.hard_reset_client_dict = collections.OrderedDict([
            ('opcode', b'\x38'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('packet_id_array_length', b'\x00'),
            ('packet_id', b'\x00\x01\x02\x03'),
        ])
        self.hard_reset_client_bytes = b''.join(self.hard_reset_client_dict.values())

        self.hard_reset_client = OpenVpnPacketHardResetClientV2(
            session_id=0x0001020304050607,
            packet_id=0x00010203,
        )

    def test_error_non_empty_packet_id_array(self):
        hard_reset_client_dict = collections.OrderedDict([
            ('opcode', b'\x38'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('packet_id_array_length', b'\x01'),
            ('packet_id_array', b'\x00\x01\x02\x03'),
            ('remote_session_id', b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
        ])
        hard_reset_client_bytes = b''.join(hard_reset_client_dict.values())
        with self.assertRaises(InvalidValue) as context_manager:
            OpenVpnPacketVariant.parse_exact_size(hard_reset_client_bytes)
        self.assertEqual(context_manager.exception.value, [0x00010203])

    def test_parse(self):
        hard_reset_client = OpenVpnPacketVariant.parse_exact_size(self.hard_reset_client_bytes)
        self.assertEqual(hard_reset_client.session_id, 0x0001020304050607)
        self.assertEqual(hard_reset_client.packet_id_array, [])
        self.assertEqual(hard_reset_client.packet_id, 0x00010203)

    def test_compose(self):
        self.assertEqual(self.hard_reset_client.compose(), self.hard_reset_client_bytes)


class TestOpenVpnPacketHardResetServerV2(unittest.TestCase):
    def setUp(self):
        self.hard_reset_server_dict = collections.OrderedDict([
            ('opcode', b'\x40'),
            ('session_id', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('packet_id_array_length', b'\x01'),
            ('packet_id_array', b'\x04\x05\x06\x07'),
            ('remote_session_id', b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'),
            ('packet_id', b'\x00\x01\x02\x03'),
        ])
        self.hard_reset_server_bytes = b''.join(self.hard_reset_server_dict.values())

        self.hard_reset_server = OpenVpnPacketHardResetServerV2(
            session_id=0x0001020304050607,
            packet_id_array=[0x04050607],
            remote_session_id=0x08090a0b0c0d0e0f,
            packet_id=0x00010203,
        )

    def test_parse(self):
        hard_reset_server = OpenVpnPacketVariant.parse_exact_size(self.hard_reset_server_bytes)
        self.assertEqual(hard_reset_server.session_id, 0x0001020304050607)
        self.assertEqual(hard_reset_server.packet_id_array, [0x04050607])
        self.assertEqual(hard_reset_server.remote_session_id, 0x08090a0b0c0d0e0f)
        self.assertEqual(hard_reset_server.packet_id, 0x00010203)

    def test_compose(self):
        self.assertEqual(self.hard_reset_server.compose(), self.hard_reset_server_bytes)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import collections
import datetime
import enum
import json

import attr
import urllib3

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import Base64Data, convert_base64_data, convert_value_to_object, convert_url

from cryptoparser.common.base import Serializable
from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.common.parse import ParserText, ParsableBase, ParsableBaseNoABC, ComposerText


class FieldParsableBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_separator(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_name(cls, parsable):
        separator = cls.get_separator()
        parser = ParserText(parsable)

        parser.parse_string_until_separator_or_end('name', separator)

        return parser

    @classmethod
    def _compose_name(cls, name):
        composer = ComposerText()

        composer.compose_string(name)

        return composer


@attr.s
class NameValueVariantBase(FieldParsableBase):
    value = attr.ib()

    @value.validator
    def _x_validator(self, attribute, value):  # pylint: disable=unused-argument
        value_class = self._get_value_class()
        if not isinstance(value, value_class):
            self.value = value_class(value)

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_separator(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_name_and_separator(cls, parsable):
        separator = cls.get_separator()
        parser = cls._parse_name(parsable)

        parser.parse_separator(separator)

        if parser['name'].lower() != cls.get_canonical_name().lower():
            raise InvalidType()

        return parser


@attr.s
class NameValuePair(FieldParsableBase):
    name = attr.ib(validator=attr.validators.instance_of(str))
    value = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
    quoted = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(bool)), default=False)

    @classmethod
    def get_separator(cls):
        return '='

    @classmethod
    def _parse(cls, parsable):
        value = None
        quoted = False

        parser = cls._parse_name(parsable)
        if parser.unparsed_length:
            parser.parse_separator(cls.get_separator())
            parser.parse_string_by_length('value', min_length=0)
            value = parser['value']
            if value and value[0] == '"':
                quoted = True
                value = value[1:]
                if value and value[-1:] == '"':
                    value = value[:-1]

        return cls(parser['name'], value, quoted), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.name)
        if self.value is not None:
            composer.compose_separator(self.get_separator())
            if self.quoted:
                composer.compose_separator('"')
            composer.compose_string(self.value)
            if self.quoted:
                composer.compose_separator('"')

        return composer.composed


@attr.s
class NameValuePairList(ParsableBase, Serializable):
    value = attr.ib(
        default=collections.OrderedDict([]),
        validator=attr.validators.instance_of(collections.OrderedDict),
    )

    @classmethod
    @abc.abstractmethod
    def get_separator(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_array(
            'value',
            cls.get_separator(),
            item_class=NameValuePair,
            separator_spaces=' \t',
            skip_empty=True
        )

        return cls(
            collections.OrderedDict([(component.name, component.value) for component in parser['value']])
        ), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        separator = self.get_separator() + ' '
        for item_number, (name, value) in enumerate(self.value.items()):
            composer.compose_string(name)
            if value is not None:
                composer.compose_separator('=')
                composer.compose_string(value)

            if item_number + 1 < len(self.value):
                composer.compose_separator(separator)

        return composer.composed

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


class NameValuePairListCommaSeparated(NameValuePairList):
    @classmethod
    def get_separator(cls):
        return ','


class NameValuePairListSemicolonSeparated(NameValuePairList):
    @classmethod
    def get_separator(cls):
        return ';'


def is_validator_optional(validator):
    return isinstance(validator, attr.validators._OptionalValidator)  # pylint: disable=protected-access


@attr.s
class FieldValueComponentBase(ParsableBase, Serializable):
    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def convert(cls, value):
        if not isinstance(value, cls):
            value = cls(value)

        return value

    @classmethod
    def _check_name_insensitive(cls, name):
        if name.lower() != cls.get_canonical_name().lower():
            raise InvalidType()

    @classmethod
    def _check_name(cls, name):
        if name != cls.get_canonical_name():
            raise InvalidType()


@attr.s
class FieldValueComponentOption(FieldValueComponentBase):
    value = attr.ib(validator=attr.validators.instance_of(bool))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _check_name(cls, name):
        cls._check_name_insensitive(name)

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        try:
            canonical_name = cls.get_canonical_name()
            parser.parse_string_by_length('value', len(canonical_name), len(canonical_name))
            cls._check_name(parser['value'])
            value = True
        except (InvalidValue, NotEnoughData):
            value = False

        return cls(value), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        if self.value:
            composer.compose_string(self.get_canonical_name())

        return composer.composed

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


@attr.s
class FieldValueComponentKeyValueBase(FieldValueComponentBase):
    value = attr.ib()

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_value(cls, parser):
        raise NotImplementedError()

    def _get_value_as_simple_type(self):
        # neccessary only because PY2 handles multiple inheritance differently than PY3
        if isinstance(self.value, ParsableBaseNoABC):
            return self.value.compose().decode('ascii')

        return self.value

    def _get_value_as_str(self):
        return str(self._get_value_as_simple_type())

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        name = cls.get_canonical_name()
        try:
            parser.parse_string_by_length('name', len(name), len(name))
        except NotEnoughData as e:
            raise InvalidType from e

        cls._check_name(parser['name'])

        if cls.get_canonical_name():
            parser.parse_separator('=')
        cls._parse_value(parser)
        parsed_value = parser['value']
        if cls.get_canonical_name():
            parsed_value = cls(parsed_value)

        return parsed_value, parser.parsed_length

    def compose(self):
        composer = ComposerText()

        if self.get_canonical_name():
            composer.compose_string_array([self.get_canonical_name(), self._get_value_as_str()], '=')
        else:
            composer.compose_string(self._get_value_as_str())

        return composer.composed

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


@attr.s
class FieldValueComponentParsableBase(FieldValueComponentKeyValueBase):
    value = attr.ib()

    def __attrs_post_init__(self):
        value_class = self._get_value_class()
        if not isinstance(self.value, value_class):
            self.value = value_class(self.value)

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_parsable('value', cls._get_value_class())


class FieldValueComponentParsable(FieldValueComponentParsableBase):
    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()


class FieldValueComponentParsableOptional(FieldValueComponentParsableBase):
    def __attrs_post_init__(self):
        if self.value is not None:
            super().__attrs_post_init__()

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()


@attr.s
class FieldValueComponentQuotedString(FieldValueComponentKeyValueBase):
    value = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    def _get_value_as_str(self):
        return f'"{self.value}"'

    def _get_value_as_simple_type(self):
        return self.value

    @classmethod
    def _parse_value(cls, parser):

        parser.parse_separator('"', 0, None)
        parser.parse_string_until_separator_or_end('value', '"')
        parser.parse_separator('"', 0, None)


@attr.s
class FieldValueComponentDateTime(FieldValueComponentKeyValueBase):
    value = attr.ib(validator=attr.validators.instance_of(datetime.datetime))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_date_time('value')

    def _get_value_as_simple_type(self):
        return self.value.strftime('%a, %d %b %Y %H:%M:%S GMT')


@attr.s
class FieldValueComponentTimeDelta(FieldValueComponentKeyValueBase):
    value = attr.ib(
        converter=convert_value_to_object(datetime.timedelta),
        validator=attr.validators.instance_of(datetime.timedelta)
    )

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def convert(cls, value):
        if isinstance(value, cls):
            return value
        if isinstance(value, datetime.timedelta):
            return cls(value)

        return cls(datetime.timedelta(seconds=value))

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_time_delta('value')

    def _get_value_as_simple_type(self):
        return int(self.value.total_seconds())

    def _as_markdown(self, level):
        return self._markdown_result(str(self.value), level)


@attr.s
class FieldValueComponentStringBase64(FieldValueComponentQuotedString):
    value = attr.ib(
        converter=convert_base64_data(),
        validator=attr.validators.instance_of(Base64Data)
    )

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _check_name(cls, name):
        cls._check_name_insensitive(name)

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_string_by_length('value')


@attr.s
class FieldValueComponentBool(FieldValueComponentKeyValueBase):
    value = attr.ib(validator=attr.validators.instance_of(bool))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    def _get_value_as_str(self):
        return 'yes' if self.value else 'no'

    def _get_value_as_simple_type(self):
        return self.value

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_bool('value')


@attr.s
class FieldValueComponentNumber(FieldValueComponentKeyValueBase):
    value = attr.ib(validator=attr.validators.instance_of(int))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_numeric('value')


@attr.s
class FieldValueComponentFloat(FieldValueComponentNumber):
    value = attr.ib(
        converter=float,
        validator=attr.validators.instance_of(float)
    )

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    def _get_value_as_simple_type(self):
        return self.value

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_float('value')


@attr.s
class FieldValueComponentPercent(FieldValueComponentNumber):
    def __attrs_post_init__(self):
        if self.value < 0 or self.value > 100:
            raise InvalidValue(self.value, type(self), 'value')

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()


@attr.s
class FieldValueComponentStringEnumParams():
    code = attr.ib(validator=attr.validators.instance_of(str))


@attr.s
class FieldValueComponentStringEnum(FieldValueComponentKeyValueBase):
    value = attr.ib()

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_type(cls):
        raise NotImplementedError()

    @value.validator
    def _validator_value(self, _, value):
        if not isinstance(value, self._get_value_type()):
            raise InvalidValue(value, type(self), 'value')

    @classmethod
    def _parse_value(cls, parser):
        try:
            parser.parse_parsable('value', cls._get_value_type())
        except InvalidValue as e:
            raise InvalidValue(e.value.decode('ascii'), cls, 'value') from e

    def _get_value_as_simple_type(self):
        return self.value.value.code


@attr.s
class FieldValueComponentStringEnumOption(FieldValueComponentStringEnum):
    @classmethod
    @abc.abstractmethod
    def _get_value_type(cls):
        raise NotImplementedError()

    @classmethod
    def get_canonical_name(cls):
        return ''

    @classmethod
    def _check_name(cls, name):
        pass


@attr.s
class FieldValueComponentString(FieldValueComponentKeyValueBase):
    value = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)))

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_string_by_length('value')


@attr.s
class FieldValueComponentUrl(FieldValueComponentKeyValueBase):
    value = attr.ib()

    @value.validator
    def _value_validate(self, _, value):
        self.value = convert_url()(value)

        if isinstance(self.value, urllib3.util.Url):
            return

        raise InvalidValue(self.value, type(self), 'value')

    @classmethod
    @abc.abstractmethod
    def get_canonical_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_value(cls, parser):
        parser.parse_string_by_length('value', item_class=convert_url())

    def _get_value_as_simple_type(self):
        if self.value.scheme == 'mailto':
            value = 'mailto:' + self.value.path[1:]
        else:
            value = str(self.value)

        return value

    def _as_markdown(self, level):
        return self._markdown_result(self._get_value_as_simple_type(), level)


class FieldValueBase(ParsableBase, Serializable):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def _get_attr_to_validator_type_dict(cls, attr_fields_dict):
        attr_to_component_name_dict = []

        for attribute in attr_fields_dict.values():
            validator = attribute.validator
            if is_validator_optional(validator):
                validator = validator.validator

            attr_to_component_name_dict.append((attribute.name, validator.type))

        return collections.OrderedDict(attr_to_component_name_dict)


class FieldsJson(FieldValueBase):
    @classmethod
    def _parse(cls, parsable):
        try:
            raw_values = json.loads(parsable.decode('ascii'), object_pairs_hook=collections.OrderedDict)
        except ValueError as e:  # json.decoder.JSONDecodeError is derived from ValueError
            raise InvalidValue(parsable.decode('ascii'), cls, 'value') from e

        attr_fields_dict = attr.fields_dict(cls)

        return cls(**{
            attribute_name: raw_values[validator_class.get_canonical_name()]
            for attribute_name, validator_class in cls._get_attr_to_validator_type_dict(attr_fields_dict).items()
            if validator_class.get_canonical_name() in raw_values
        }), len(parsable)

    def compose(self):
        attr_fields_dict = attr.fields_dict(type(self))

        return json.dumps(collections.OrderedDict([
            (
                validator_class.get_canonical_name(),
                getattr(self, attribute_name)._get_value_as_simple_type()  # pylint: disable=protected-access
            )
            for attribute_name, validator_class in self._get_attr_to_validator_type_dict(attr_fields_dict).items()
            if getattr(self, attribute_name) is not None
        ])).encode('ascii')


class FieldValueMultiple(FieldValueBase):
    @classmethod
    @abc.abstractmethod
    def _get_header_value_list_class(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_basic_params(cls, attr_to_component_name_dict, attr_fields_dict, components, params):
        for name, attribute in attr_fields_dict.items():
            for component in components:
                try:
                    attr_to_component_name_dict[name]._check_name(component)  # pylint: disable=protected-access
                except InvalidType:
                    pass
                else:
                    components[attr_to_component_name_dict[name].get_canonical_name()] = components.pop(component)
                    break
            else:
                if attribute.default == attr.NOTHING:
                    raise InvalidValue(None, cls, name)

                component = attribute.default

            if attr_to_component_name_dict[name].get_canonical_name() in components:
                parsable = components.pop(attr_to_component_name_dict[name].get_canonical_name())
                if parsable is None:  # value is None in case of optional values
                    parsable = component
                else:
                    parsable = '='.join([attr_to_component_name_dict[name].get_canonical_name(), parsable])
                params[name] = attr_to_component_name_dict[name].parse_exact_size(parsable.encode('ascii'))
            else:
                params[name] = attribute.default

    @classmethod
    def _parse_extensions(cls, attr_to_component_name_dict, extension, components, params):
        if extension and components:
            name, _ = extension
            params[name] = attr_to_component_name_dict[name](components)

    @classmethod
    def _parse(cls, parsable):
        params = {}
        extension = None
        attr_fields_dict_basic = {}
        attr_fields_dict = attr.fields_dict(cls)
        for name, attribute in attr_fields_dict.items():
            if not attribute.metadata.get('extension', False):
                attr_fields_dict_basic[name] = attribute
            elif extension is None:
                extension = (name, attribute)
            else:
                raise NotImplementedError()
        attr_to_component_name_dict = cls._get_attr_to_validator_type_dict(attr_fields_dict)

        components = cls._get_header_value_list_class().parse_exact_size(parsable).value

        cls._parse_basic_params(attr_to_component_name_dict, attr_fields_dict_basic, components, params)
        cls._parse_extensions(attr_to_component_name_dict, extension, components, params)

        return cls(**params), len(parsable)

    def compose(self):
        composer = ComposerText()

        cls = type(self)
        attr_fields_dict = attr.fields_dict(cls)
        components = []
        for name, attribute in attr_fields_dict.items():
            field_value = getattr(self, name)
            validator = attribute.validator

            if is_validator_optional(validator):
                if field_value is None:
                    continue

                validator = validator.validator

            value = field_value.value
            if issubclass(validator.type, FieldValueComponentOption):
                if value is False:
                    continue

            components.append(getattr(self, name))

        separator = self._get_header_value_list_class().get_separator() + ' '
        composer.compose_string_array(components, separator)

        return composer.composed


class FieldsCommaSeparated(FieldValueMultiple):
    @classmethod
    def _get_header_value_list_class(cls):
        return NameValuePairListCommaSeparated


class FieldsSemicolonSeparated(FieldValueMultiple):
    @classmethod
    def _get_header_value_list_class(cls):
        return NameValuePairListSemicolonSeparated


class MimeTypeRegistry(enum.Enum):
    APPLICATION = 'application'
    AUDIO = 'audio'
    FONT = 'font'
    EXAMPLE = 'example'
    IMAGE = 'image'
    MESSAGE = 'message'
    MODEL = 'model'
    MULTIPART = 'multipart'
    TEXT = 'text'
    VIDEO = 'video'


@attr.s
class FieldValueMimeType(FieldValueComponentBase):
    type = attr.ib(
        validator=attr.validators.instance_of(str),
        default=None,
    )
    registry = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(MimeTypeRegistry)),
        default=None,
    )

    def __str__(self):
        return f'{self.registry.value}/{self.type}'

    @property
    def value(self):
        return self

    @classmethod
    def get_canonical_name(cls):
        return ''

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator('registry', '/', item_class=MimeTypeRegistry)
        parser.parse_separator('/')
        parser.parse_string_by_length('type', parser.unparsed_length)

        return FieldValueMimeType(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(str(self))

        return composer.composed

    @classmethod
    def _check_name(cls, name):
        pass


@attr.s
class FieldValueSingleBase(FieldValueBase, Serializable):
    value = attr.ib()

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_type(cls):
        raise NotImplementedError()

    @classmethod
    def convert(cls, value):
        if not isinstance(value, cls._get_value_type()):
            return value

        return cls(value)

    @value.validator
    def _value_validate(self, _, value):
        value_type = self._get_value_type()
        if not isinstance(value, value_type):
            raise InvalidValue(value, value_type, 'value')

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


class FieldValueSingleSimpleBase(FieldValueSingleBase):
    @classmethod
    @abc.abstractmethod
    def _value_from_str(cls, value):
        raise NotImplementedError()


class FieldValueSingle(FieldValueSingleSimpleBase):
    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_by_length('value')
        value = cls._value_from_str(parser['value'])

        return cls(value), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value)

        return composer.composed

    @classmethod
    @abc.abstractmethod
    def _get_value_type(cls):
        raise NotImplementedError()


@attr.s
class FieldValueStringEnumParams(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(str))
    human_readable_name = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str))
    )

    def _as_markdown(self, level):
        if self.human_readable_name:
            return self._markdown_result(self.human_readable_name, level)

        return False, self.code.replace('_', ' ')


class FieldValueString(FieldValueSingle):
    @classmethod
    def _get_value_type(cls):
        return str

    @classmethod
    def _value_from_str(cls, value):
        return str(value)


class FieldValueSingleComplexBase(FieldValueSingleBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class FieldValueDateTime(FieldValueSingleComplexBase):
    @classmethod
    def _get_value_type(cls):
        return datetime.datetime

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_date_time('value')

        return cls(parser['value']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_date_time(self.value, '%a, %d %b %Y %H:%M:%S GMT')

        return composer.composed

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


class FieldValueStringBySeparatorBase(FieldValueSingleComplexBase):
    @classmethod
    @abc.abstractmethod
    def _get_separators(cls):
        raise NotImplementedError()

    @classmethod
    def _get_value_type(cls):
        return str

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator_or_end('value', cls._get_separators())

        return cls(parser['value']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value)

        return composer.composed


class FieldValueStringEnum(FieldValueSingleComplexBase):
    @classmethod
    @abc.abstractmethod
    def _get_value_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        try:
            value = cls._get_value_type().parse_exact_size(parsable)
        except InvalidValue as e:
            raise InvalidValue(parsable.decode('ascii'), cls, 'value') from e

        return cls(value), len(parsable)

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value.value.code)

        return composer.composed


class FieldValueTimeDelta(FieldValueSingleComplexBase):
    @classmethod
    def _get_value_type(cls):
        return datetime.timedelta

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_time_delta('value')

        return cls(parser['value']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_time_delta(self.value)

        return composer.composed



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections

import unittest

import ipaddress

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import InvalidType
from cryptoparser.common.field import NameValuePairListSemicolonSeparated
from cryptoparser.dnsrec.txt import (
    DmarcAlignment,
    DmarcFailureReportingFormat,
    DmarcFailureReportingOption,
    DmarcPolicyOption,
    DmarcPolicyVersion,
    DmarcReportingInterval,
    DnsRecordTxtValueDmarc,
    DnsRecordTxtValueMtaSts,
    DnsRecordTxtValueSpf,
    DnsRecordTxtValueSpfDirectiveA,
    DnsRecordTxtValueSpfDirectiveAll,
    DnsRecordTxtValueSpfDirectiveExists,
    DnsRecordTxtValueSpfDirectiveInclude,
    DnsRecordTxtValueSpfDirectiveIp4,
    DnsRecordTxtValueSpfDirectiveIp6,
    DnsRecordTxtValueSpfDirectiveMx,
    DnsRecordTxtValueSpfDirectivePtr,
    DnsRecordTxtValueSpfModifierExplanation,
    DnsRecordTxtValueSpfModifierRedirect,
    DnsRecordTxtValueSpfModifierUnknown,
    DnsRecordTxtValueTlsRpt,
    MtaStsPolicyVersion,
    SpfQualifier,
    SpfVersion,
    TlsRptVersion,
)


class TestDnsRecordTxtValueDmarcReportingInterval(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            DmarcReportingInterval.parse_exact_size(b'ri=5000000000')

        self.assertEqual(context_manager.exception.value, 5000000000)


class TestDnsRecordDmarc(unittest.TestCase):
    _record_minimal = DnsRecordTxtValueDmarc(DmarcPolicyVersion.DMARC1, DmarcPolicyOption.NONE)
    _record_minimal_bytes = b'v=DMARC1;p=none'
    _record_full = DnsRecordTxtValueDmarc(
        version=DmarcPolicyVersion.DMARC1,
        policy=DmarcPolicyOption.NONE,
        alignment_dkim=DmarcAlignment.STRICT,
        alignment_aspf=DmarcAlignment.STRICT,
        failure_option=DmarcFailureReportingOption.ANY_FAILURE,
        percent=99,
        reporting_url_aggregated='mailto:dmarc-report@example.com',
        reporting_url_failure='https://example.com/dmarc/report/failure',
        reporting_format=DmarcFailureReportingFormat.AUTHENTICATION_FAILURE_REPORTING_FORMAT,
        reporting_interval=3600,
        subdomain_policy=DmarcPolicyOption.NONE,
    )
    _record_full_bytes = b'; '.join([
        b'v=DMARC1',
        b'p=none',
        b'adkim=s',
        b'aspf=s',
        b'fo=1',
        b'pct=99',
        b'rua=mailto:dmarc-report@example.com',
        b'ruf=https://example.com/dmarc/report/failure',
        b'rf=afrf',
        b'ri=3600',
        b'sp=none',
    ])

    def test_parse(self):
        self.assertEqual(DnsRecordTxtValueDmarc.parse_exact_size(self._record_minimal_bytes), self._record_minimal)
        self.assertEqual(DnsRecordTxtValueDmarc.parse_exact_size(self._record_full_bytes), self._record_full)

    def test_compose(self):
        self.assertEqual(self._record_full.compose(), self._record_full_bytes)


class TestDnsRecordMtaSts(unittest.TestCase):
    _record_minimal = DnsRecordTxtValueMtaSts(MtaStsPolicyVersion.STSV1, '20160831085700Z')
    _record_minimal_bytes = b'v=STSv1; id=20160831085700Z'
    _record_full = DnsRecordTxtValueMtaSts(
        version=MtaStsPolicyVersion.STSV1,
        identifier='20160831085700Z',
        extensions=NameValuePairListSemicolonSeparated(
            collections.OrderedDict([('extension_name', 'extension_value')])
        ),
    )
    _record_full_bytes = b'v=STSv1; id=20160831085700Z; extension_name=extension_value'

    def test_parse(self):
        self.assertEqual(DnsRecordTxtValueMtaSts.parse_exact_size(self._record_minimal_bytes), self._record_minimal)
        self.assertEqual(DnsRecordTxtValueMtaSts.parse_exact_size(self._record_full_bytes), self._record_full)

    def test_compose(self):
        self.assertEqual(self._record_minimal.compose(), self._record_minimal_bytes)
        self.assertEqual(self._record_full.compose(), self._record_full_bytes)


class TestDnsRecordTxtValueTlsRpt(unittest.TestCase):
    _record_minimal = DnsRecordTxtValueTlsRpt(TlsRptVersion.TLSRPTV1, 'https://example.com/tlsrpt/report/failure')
    _record_minimal_bytes = b'v=TLSRPTv1; rua=https://example.com/tlsrpt/report/failure'
    _record_full = DnsRecordTxtValueTlsRpt(
        version=TlsRptVersion.TLSRPTV1,
        reporting_url_aggregated='mailto:tls-report@example.com',
        extensions=NameValuePairListSemicolonSeparated(
            collections.OrderedDict([('extension_name', 'extension_value')])
        ),
    )
    _record_full_bytes = b'v=TLSRPTv1; rua=mailto:tls-report@example.com; extension_name=extension_value'

    def test_parse(self):
        self.assertEqual(DnsRecordTxtValueTlsRpt.parse_exact_size(self._record_minimal_bytes), self._record_minimal)
        self.assertEqual(DnsRecordTxtValueTlsRpt.parse_exact_size(self._record_full_bytes), self._record_full)

    def test_compose(self):
        self.assertEqual(self._record_minimal.compose(), self._record_minimal_bytes)
        self.assertEqual(self._record_full.compose(), self._record_full_bytes)


class TestDnsRecordTxtValueSpfDirective(unittest.TestCase):
    def test_parse_domain_optional(self):
        directive = DnsRecordTxtValueSpfDirectivePtr.parse_exact_size(b'ptr')
        self.assertEqual(directive.domain, None)
        self.assertEqual(directive.compose(), b'ptr')

        directive = DnsRecordTxtValueSpfDirectivePtr.parse_exact_size(b'ptr:domain')
        self.assertEqual(directive.domain.value, 'domain')
        self.assertEqual(directive.compose(), b'ptr:domain')

    def test_parse_domain_required(self):
        directive = DnsRecordTxtValueSpfDirectiveInclude.parse_exact_size(b'include:domain')
        self.assertEqual(directive.domain.value, 'domain')
        self.assertEqual(directive.compose(), b'include:domain')

    def test_parse_single_cidr_length(self):
        directive = DnsRecordTxtValueSpfDirectiveIp4.parse_exact_size(b'ip4:1.1.1.1')
        self.assertEqual(directive.ipv4_network, ipaddress.IPv4Network('1.1.1.1/32'))
        self.assertEqual(directive.compose(), b'ip4:1.1.1.1')

        directive = DnsRecordTxtValueSpfDirectiveIp4.parse_exact_size(b'ip4:1.1.1.0/24')
        self.assertEqual(directive.ipv4_network, ipaddress.IPv4Network('1.1.1.0/24'))
        self.assertEqual(directive.compose(), b'ip4:1.1.1.0/24')

        directive = DnsRecordTxtValueSpfDirectiveIp6.parse_exact_size(b'ip6:::1')
        self.assertEqual(directive.ipv6_network, ipaddress.IPv6Network('::1/128'))
        self.assertEqual(directive.compose(), b'ip6:::1')

        directive = DnsRecordTxtValueSpfDirectiveIp6.parse_exact_size(b'ip6:::1:0/120')
        self.assertEqual(directive.ipv6_network, ipaddress.IPv6Network('::1:0/120'))
        self.assertEqual(directive.compose(), b'ip6:::1:0/120')

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordTxtValueSpfDirectiveMx('example.com', ipv4_cidr_length=-1)
        self.assertEqual(context_manager.exception.value, -1)

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordTxtValueSpfDirectiveMx('example.com', ipv4_cidr_length=33)
        self.assertEqual(context_manager.exception.value, 33)

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordTxtValueSpfDirectiveMx('example.com', ipv6_cidr_length=-1)
        self.assertEqual(context_manager.exception.value, -1)

        with self.assertRaises(InvalidValue) as context_manager:
            DnsRecordTxtValueSpfDirectiveMx('example.com', ipv6_cidr_length=129)
        self.assertEqual(context_manager.exception.value, 129)

    def test_parse_dual_cidr_length(self):
        directive = DnsRecordTxtValueSpfDirectiveMx.parse_exact_size(b'mx:example.com')
        self.assertEqual(directive.domain.value, 'example.com')
        self.assertEqual(directive.compose(), b'mx:example.com')

        directive = DnsRecordTxtValueSpfDirectiveMx.parse_exact_size(b'mx:example.com/24')
        self.assertEqual(directive.domain.value, 'example.com')
        self.assertEqual(directive.ipv4_cidr_length, 24)
        self.assertEqual(directive.ipv6_cidr_length, None)
        self.assertEqual(directive.compose(), b'mx:example.com/24')

        directive = DnsRecordTxtValueSpfDirectiveMx.parse_exact_size(b'mx:example.com/24/64')
        self.assertEqual(directive.domain.value, 'example.com')
        self.assertEqual(directive.ipv4_cidr_length, 24)
        self.assertEqual(directive.ipv6_cidr_length, 64)
        self.assertEqual(directive.compose(), b'mx:example.com/24/64')


class TestDnsRecordTxtValueSpf(unittest.TestCase):
    _record_minimal = DnsRecordTxtValueSpf(version=SpfVersion.SPF1, terms=[])
    _record_minimal_bytes = b'v=spf1'
    _record_full = DnsRecordTxtValueSpf(
        version=SpfVersion.SPF1,
        terms=[
            DnsRecordTxtValueSpfDirectiveExists(domain='%{ir}.%{l1r+-}._spf.%{d}'),
            DnsRecordTxtValueSpfDirectiveA('example.com', 32, 128),
            DnsRecordTxtValueSpfDirectiveMx('example.com', 32, 128),
            DnsRecordTxtValueSpfDirectivePtr('example.com'),
            DnsRecordTxtValueSpfDirectiveIp4('1.2.3.4'),
            DnsRecordTxtValueSpfDirectiveIp6('::1:2:3:4'),
            DnsRecordTxtValueSpfDirectiveInclude('_spf.example.com'),
            DnsRecordTxtValueSpfModifierRedirect('redirect.example.com'),
            DnsRecordTxtValueSpfModifierExplanation('exp.example.com'),
            DnsRecordTxtValueSpfModifierUnknown('modifier_key', 'modifier_value'),
            DnsRecordTxtValueSpfDirectiveAll(SpfQualifier.FAIL),
        ],
    )
    _record_full_bytes = b' '.join([
        b'v=spf1',
        b'exists:%{ir}.%{l1r+-}._spf.%{d}',
        b'a:example.com/32/128',
        b'mx:example.com/32/128',
        b'ptr:example.com',
        b'ip4:1.2.3.4',
        b'ip6:::1:2:3:4',
        b'include:_spf.example.com',
        b'redirect=redirect.example.com',
        b'exp=exp.example.com',
        b'modifier_key=modifier_value',
        b'-all',
    ])

    def test_error_non_spf(self):
        with self.assertRaises(InvalidType):
            DnsRecordTxtValueSpf.parse_exact_size(b'v=STSv1')

    def test_parse(self):
        self.assertEqual(DnsRecordTxtValueSpf.parse_exact_size(self._record_minimal_bytes), self._record_minimal)
        self.assertEqual(DnsRecordTxtValueSpf.parse_exact_size(self._record_full_bytes), self._record_full)

    def test_compose(self):
        self.assertEqual(self._record_minimal.compose(), self._record_minimal_bytes)
        self.assertEqual(self._record_full.compose(), self._record_full_bytes)

    def test_as_markdown(self):
        self.assertEqual(self._record_minimal.as_markdown(), '\n'.join([
            '* Version: SPF1',
            '* Terms: -',
            '',
        ]))

        self.assertEqual(self._record_full.as_markdown(), '\n'.join([
            '* Version: SPF1',
            '* Terms:',
            '    * Exists:',
            '        * Domain: %{ir}.%{l1r+-}._spf.%{d}',
            '        * Qualifier: n/a',
            '    * A/AAAA records:',
            '        * Domain: example.com',
            '        * Ipv4 Cidr Length: 32',
            '        * Ipv6 Cidr Length: 128',
            '        * Qualifier: n/a',
            '    * MX records:',
            '        * Domain: example.com',
            '        * Ipv4 Cidr Length: 32',
            '        * Ipv6 Cidr Length: 128',
            '        * Qualifier: n/a',
            '    * PTR records:',
            '        * Domain: example.com',
            '        * Qualifier: n/a',
            '    * IPv4 records:',
            '        * Ipv4 Network: 1.2.3.4/32',
            '        * Qualifier: n/a',
            '    * IPv6 records:',
            '        * Ipv6 Network: ::1:2:3:4/128',
            '        * Qualifier: n/a',
            '    * Include:',
            '        * Domain: _spf.example.com',
            '        * Qualifier: n/a',
            '    * Redirect: redirect.example.com',
            '    * Explanation: exp.example.com',
            '    * Modifier Key: modifier_value',
            '    * All:',
            '        * Qualifier: Fail',
            '',
        ]))
